"""Perform queries against the cross references (xref) endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _CrossReferences(object):
    """Perform CrossReferences queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_xrefs_symbol(self, species, symbol, db_type=None, external_db=None,
                         object_type=None, norm_func=brc.default_pandas):
        """
        Looks up an external symbol and returns all Ensembl objects linked to
        it. This can be a display name for a gene/transcript/translation, a
        synonym or an externally linked reference. If a gene's transcript is
        linked to the supplied symbol the service will return both gene and
        transcript (it supports transient links). GET method. Queries:
        `xrefs/symbol/:species/:symbol <https://rest.ensembl.org/documentation/info/xref_external>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        symbol : str
            Symbol or display name of a gene. Example values: 'BRCA2'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        external_db : str or NoneType, optional, default: NoneType
            Filter by external database. Example values: 'HGNC'.
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene'.'transcript'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'xrefs/symbol/{species}/{symbol}'.format(
            species=species,
            symbol=symbol
        )

   
        return self.rest_query(
                method_type,
                endpoint,
                db_type=db_type,
                external_db=external_db,
                object_type=object_type,
                norm_func=norm_func
            )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_xrefs_id(self, query_id, all_levels=None, db_type=None,
                     external_db=None, object_type=None, species=None,
                     norm_func=brc.default_pandas):
        """
        Perform lookups of Ensembl Identifiers and retrieve their external
        references in other databases. GET method. Queries:
        `xrefs/id/:id <https://rest.ensembl.org/documentation/info/xref_id>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl Stable ID. Example values: 'ENSG00000157764'.
        all_levels : bool or NoneType, optional, default: NoneType
            Set to find all genetic features linked to the stable ID, and fetch
            all external references for them. Specifying this on a gene will
            also return values from its transcripts and translations. Example
            values: '1'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        external_db : str or NoneType, optional, default: NoneType
            Filter by external database. Example values: 'HGNC'.
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene transcript'.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data

        Notes
        -----
        When the external_db is included this seems to return no data. I am not
        sure if this is a fault with this function or the Ensembl REST endpoint
        , I will need to research this.
        """
        method_type = 'GET'
        endpoint = 'xrefs/id/{id}'.format(
            id=query_id
        )

        if external_db is not None:
            all_levels = '1'

        return self.rest_query(
            method_type,
            endpoint,
            all_levels=all_levels,
            db_type=db_type,
            external_db=external_db,
            object_type=object_type,
            species=species,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_xrefs_name(self, name, species, db_type=None, external_db=None,
                       norm_func=brc.default_pandas):
        """
        Performs a lookup based upon the primary accession or display label of
        an external reference and returning the information held by Ensembl
        about the entry. GET method. Queries:
        `xrefs/name/:species/:name <https://rest.ensembl.org/documentation/info/xref_name>`_.

        Parameters
        ----------
        name : str
            Symbol or display name of a gene. Example values: 'BRCA2'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        external_db : str or NoneType, optional, default: NoneType
            Filter by external database. Example values: 'HGNC'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data

        
        """
        method_type = 'GET'
        endpoint = 'xrefs/name/{species}/{name}'.format(
            species=species,
            name=name
        )

  
        return self.rest_query(
                method_type,
                endpoint,
                db_type=db_type,
                external_db=external_db,
                norm_func=norm_func
                )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CrossReferences(brc._BaseRestClient, _CrossReferences):
    """Perform CrossReferences queries
    """
    pass



"""Perform queries against the transcript haplotypes endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _TranscriptHaplotypes(object):
    """Perform TranscriptHaplotypes queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_transcript_haplotypes(self, query_id, species,
                                  aligned_sequences=None, samples=None,
                                  sequence=None, norm_func=brc.default_pandas):
        """
        Computes observed transcript haplotype sequences based on phased
        genotype data. GET method. Queries:
        `transcript_haplotypes/:species/:id <https://rest.ensembl.org/documentation/info/transcript_haplotypes_get>`_.

        Parameters
        ----------
        query_id : str
            Transcript stable id. Example values: 'ENST00000288602'.
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        aligned_sequences : bool or NoneType, optional, default: NoneType
            Include aligned sequences used to generate differences. Example
            values: '1'.
        samples : bool or NoneType, optional, default: NoneType
            Include sample-haplotype assignments. Example values: '1'.
        sequence : bool or NoneType, optional, default: NoneType
            Include raw sequences. Example values: '1'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'transcript_haplotypes/{species}/{id}'.format(
            species=species,
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned_sequences=aligned_sequences,
            samples=samples,
            sequence=sequence,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class TranscriptHaplotypes(brc._BaseRestClient, _TranscriptHaplotypes):
    """Perform TranscriptHaplotypes queries
    """
    pass



"""Perform queries against the variation GA4GH endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _VariationGa4Gh(object):
    """Perform VariationGa4Gh queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_beacon(self, norm_func=brc.default_pandas):
        """
        Return Beacon information. GET method. Queries:
        `ga4gh/beacon <https://rest.ensembl.org/documentation/info/beacon_get>`_.

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/beacon'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_beacon_query(self, alternate_bases, assembly_id, end,
                               end_max, end_min, reference_bases,
                               reference_name, start, start_max, start_min,
                               variant_type, dataset_ids=None,
                               include_dataset_responses=None,
                               norm_func=brc.default_pandas):
        """
        Return the Beacon response for allele information. GET method. Queries:
        `ga4gh/beacon/query <https://rest.ensembl.org/documentation/info/beacon_query_get>`_.

        Parameters
        ----------
        alternate_bases : str
            The bases that appear instead of the reference bases. Accepted
            values: see the ALT field in VCF 4.2 specification
            (http://samtools.github.io/hts-specs/VCFv4.2.pdf) Note: either
            alternateBases or variantType is required. Example values: 'C'.
        assembly_id : str
            Assembly identifier (GRC notation, e.g. GRCh38). Example values:
            'GRCh38'.
        end : int
            Precise end position, allele locus (0-based). Accepted values: non-
            negative integers smaller than reference length. Note: only
            required when using variantType to query precise structural
            variants. Example values: '__VAR(GA4GH_beacon_end)__'.
        end_max : int
            Maximum end coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_endmax)__'.
        end_min : int
            Minimum end coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_endmin)__'.
        reference_bases : str
            Reference bases for this variant (starting from start). Accepted
            values: see the REF field in VCF 4.2 specification
            (http://samtools.github.io/hts-specs/VCFv4.2.pdf). Example values:
            'G'.
        reference_name : str
            Reference name (chromosome). Accepted values: 1-22, X, Y, MT.
            Example values: '9'.
        start : int
            Precise start position, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: Do not
            use when querying for imprecise structural variants, use startMin
            and startMax instead. Example values: '22125503'.
        start_max : int
            Maximum start coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_startmax)__'.
        start_min : int
            Minimum start coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_startmin)__'.
        variant_type : str
            Used to denote structural variants. Accepted values:
            DUP,DEL,INS,INV,CNV,DUP:TANDEM, see the ALT field in VCF 4.2
            specification (http://samtools.github.io/hts-specs/VCFv4.2.pdf)
            Note: either alternateBases or variantType is required. Example
            values: '__VAR(GA4GH_beacon_variantType)__'.
        dataset_ids : array of strings or NoneType, optional, default: NoneType
            Identifiers of datasets. Identifiers have to be chosen from 'Short
            name' column in the Variant sets list
            (http://www.ensembl.org/info/genome/variation/species/sets.html).
        include_dataset_responses : str or NoneType, optional, default: NoneType
            Indicator of whether responses for individual datasets should be
            included. Accepted values: ALL, HIT, MISS, NONE.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/beacon/query'

        return self.rest_query(
            method_type,
            endpoint,
            alternateBases=alternate_bases,
            assemblyId=assembly_id,
            end=end,
            endMax=end_max,
            endMin=end_min,
            referenceBases=reference_bases,
            referenceName=reference_name,
            start=start,
            startMax=start_max,
            startMin=start_min,
            variantType=variant_type,
            datasetIds=dataset_ids,
            includeDatasetResponses=include_dataset_responses,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_beacon_query(self, alternate_bases, assembly_id, end,
                                end_max, end_min, reference_bases,
                                reference_name, start, start_max, start_min,
                                variant_type, dataset_ids=None,
                                include_dataset_responses=None,
                                norm_func=brc.default_pandas):
        """
        Return the Beacon response for allele information. POST method. Queries:
        `ga4gh/beacon/query <https://rest.ensembl.org/documentation/info/beacon_query_post>`_.

        Parameters
        ----------
        alternate_bases : str
            The bases that appear instead of the reference bases. Accepted
            values: see the ALT field in VCF 4.2 specification
            (http://samtools.github.io/hts-specs/VCFv4.2.pdf) Note: either
            alternateBases or variantType is required. Example values: 'C'.
        assembly_id : str
            Assembly identifier (GRC notation, e.g. GRCh38). Example values:
            'GRCh38'.
        end : int
            Precise end position, allele locus (0-based). Accepted values: non-
            negative integers smaller than reference length. Note: only
            required when using variantType to query precise structural
            variants. Example values: '__VAR(GA4GH_beacon_end)__'.
        end_max : int
            Maximum end coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_endmax)__'.
        end_min : int
            Minimum end coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_endmin)__'.
        reference_bases : str
            Reference bases for this variant (starting from start). Accepted
            values: see the REF field in VCF 4.2 specification
            (http://samtools.github.io/hts-specs/VCFv4.2.pdf). Example values:
            'G'.
        reference_name : str
            Reference name (chromosome). Accepted values: 1-22, X, Y, MT.
            Example values: '9'.
        start : int
            Precise start position, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: Do not
            use when querying for imprecise structural variants, use startMin.
            Example values: '22125503'.
        start_max : int
            Maximum start coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_startmax)__'.
        start_min : int
            Minimum start coordinate, allele locus (0-based). Accepted values:
            non-negative integers smaller than reference length. Note: only
            required when using variantType to query imprecise structural
            variants. Example values: '__VAR(GA4GH_beacon_startmin)__'.
        variant_type : str
            Used to denote structural variants. Accepted values:
            DUP,DEL,INS,INV,CNV,DUP:TANDEM, see the ALT field in VCF 4.2
            specification (http://samtools.github.io/hts-specs/VCFv4.2.pdf)
            Note: either alternateBases or variantType is required. Example
            values: '__VAR(GA4GH_beacon_variantType)__'.
        dataset_ids : array of strings or NoneType, optional, default: NoneType
            Identifiers of datasets. Identifiers have to be chosen from 'Short
            name' column in the Variant sets list
            (http://www.ensembl.org/info/genome/variation/species/sets.html).
        include_dataset_responses : str or NoneType, optional, default: NoneType
            Indicator of whether responses for individual datasets should be
            included. Accepted values: ALL, HIT, MISS, NONE.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/beacon/query'

        return self.rest_query(
            method_type,
            endpoint,
            alternateBases=alternate_bases,
            assemblyId=assembly_id,
            end=end,
            endMax=end_max,
            endMin=end_min,
            referenceBases=reference_bases,
            referenceName=reference_name,
            start=start,
            startMax=start_max,
            startMin=start_min,
            variantType=variant_type,
            datasetIds=dataset_ids,
            includeDatasetResponses=include_dataset_responses,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_features(self, query_id, norm_func=brc.default_pandas):
        """
        Return the GA4GH record for a specific sequence feature given its
        identifier. GET method. Queries:
        `ga4gh/features/:id <https://rest.ensembl.org/documentation/info/features_id>`_.

        Parameters
        ----------
        query_id : str
            Feature id. Example values: 'ENST00000408937.7'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/features/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_features_search(self, end, reference_name, start,
                                   feature_set_id, feature_types=None,
                                   featureset_id=None, page_size=None,
                                   page_token=None, parent_id=None,
                                   norm_func=brc.default_pandas):
        """
        Return a list of sequence annotation features in GA4GH format. POST
        method. Queries:
        `ga4gh/features/search <https://rest.ensembl.org/documentation/info/features_post>`_.

        Parameters
        ----------
        end : int
            Return features within a window with this end location. Example
            values: '1109000'.
        reference_name : str
            Return features on this reference. Example values: '6'.
        start : int
            Return features within a window with this start location. Example
            values: '1108000'.
        feature_set_id : str
            Post option not defined in required/optional args.
        feature_types : array of strings or NoneType, optional, default: NoneType
            Return features of this type (requires SO terms). Example values:
            '[transcript]'.
        featureset_id : str or NoneType, optional, default: NoneType
            Return features in this set. Example values: 'Ensembl'.
        page_size : int or NoneType, optional, default: NoneType
            Number of features to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        parent_id : str or NoneType, optional, default: NoneType
            Return the child features of this feature. Example values:
            'ENSG00000176515.1'.
        """
        method_type = 'POST'
        endpoint = 'ga4gh/features/search'

        return self.rest_query(
            method_type,
            endpoint,
            end=end,
            referenceName=reference_name,
            start=start,
            featureSetId=feature_set_id,
            featureTypes=feature_types,
            featuresetId=featureset_id,
            pageSize=page_size,
            pageToken=page_token,
            parentId=parent_id,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_callsets_search(self, variant_set_id, name=None,
                                   page_size=None, page_token=None,
                                   norm_func=brc.default_pandas):
        """
        Return a list of sets of genotype calls for specific samples in GA4GH
        format. POST method. Queries:
        `ga4gh/callsets/search <https://rest.ensembl.org/documentation/info/gacallSet>`_.

        Parameters
        ----------
        variant_set_id : str
            Return callSets for a specific variant set. Example values: '1'.
        name : str or NoneType, optional, default: NoneType
            Return callSets by name. Example values: 'NA19777'.
        page_size : int or NoneType, optional, default: NoneType
            Number of callSets to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/callsets/search'

        return self.rest_query(
            method_type,
            endpoint,
            variantSetId=variant_set_id,
            name=name,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_callsets(self, query_id, norm_func=brc.default_pandas):
        """
        Return the GA4GH record for a specific CallSet given its identifier.
        GET method. Queries:
        `ga4gh/callsets/:id <https://rest.ensembl.org/documentation/info/gacallset_id>`_.

        Parameters
        ----------
        query_id : str
            CallSet id. Example values: '1'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/callsets/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_datasets_search(self, page_size=None, page_token=None,
                                   norm_func=brc.default_pandas):
        """
        Return a list of datasets in GA4GH format. POST method. Queries:
        `ga4gh/datasets/search <https://rest.ensembl.org/documentation/info/gadataset>`_.

        Parameters
        ----------
        page_size : int or NoneType, optional, default: NoneType
            Number of dataSets to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/datasets/search'

        return self.rest_query(
            method_type,
            endpoint,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_datasets(self, query_id, norm_func=brc.default_pandas):
        """
        Return the GA4GH record for a specific dataset given its identifier.
        GET method. Queries:
        `ga4gh/datasets/:id <https://rest.ensembl.org/documentation/info/gadataset_id>`_.

        Parameters
        ----------
        query_id : str
            Dataset id. Example values: '6e340c4d1e333c7a676b1710d2e3953c'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/datasets/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_featuresets_search(self, dataset_id, page_size=None,
                                      page_token=None,
                                      norm_func=brc.default_pandas):
        """
        Return a list of feature sets in GA4GH format. POST method. Queries:
        `ga4gh/featuresets/search <https://rest.ensembl.org/documentation/info/gafeatureset>`_.

        Parameters
        ----------
        dataset_id : str
            Return featureSets by dataSet Identifier. Example values:
            'Ensembl'.
        page_size : int or NoneType, optional, default: NoneType
            Number of featureSets to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/featuresets/search'

        return self.rest_query(
            method_type,
            endpoint,
            datasetId=dataset_id,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_featuresets(self, query_id, norm_func=brc.default_pandas):
        """
        Return the GA4GH record for a specific featureSet given its identifier.
        GET method. Queries:
        `ga4gh/featuresets/:id <https://rest.ensembl.org/documentation/info/gafeatureset_id>`_.

        Parameters
        ----------
        query_id : str
            featureSet id. Example values: 'Ensembl'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/featuresets/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_variants(self, query_id, norm_func=brc.default_pandas):
        """
        Return the GA4GH record for a specific variant given its identifier.
        GET method. Queries:
        `ga4gh/variants/:id <https://rest.ensembl.org/documentation/info/gavariant_id>`_.

        Parameters
        ----------
        query_id : str
            Variation id. Example values: '1:rs1333049'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/variants/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_variant_annotations_search(self, variant_annotation_set_id,
                                              effects=None, end=None,
                                              page_size=None, page_token=None,
                                              reference_id=None,
                                              reference_name=None, start=None,
                                              norm_func=brc.default_pandas):
        """
        Return variant annotation information in GA4GH format for a region on a
        reference sequence. POST method. Queries:
        `ga4gh/variantannotations/search <https://rest.ensembl.org/documentation/info/gavariantannotations>`_.

        Parameters
        ----------
        variant_annotation_set_id : str
            Return variant annotations data for a specific annotationSet
            (Ensembl is an alias for the current default annotation set).
            Example values: 'Ensembl'.
        effects : [ontologyterm] or NoneType, optional, default: NoneType
            Return variant annotation filtering on effects. Example values: '[
            {"sourceName":"SO", "term":"missense_variant", "id":"SO:0001583",
            "sourceVersion": ""}]'.
        end : int or NoneType, optional, default: NoneType
            End position of region (zero-based, exclusive). Example values:
            '25221500'.
        page_size : int or NoneType, optional, default: NoneType
            Number of variants to show per page.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        reference_id : str or NoneType, optional, default: NoneType
            Reference sequence id. Example values:
            '9489ae7581e14efcad134f02afafe26c'.
        reference_name : str or NoneType, optional, default: NoneType
            Reference sequence name. Example values: '22'.
        start : int or NoneType, optional, default: NoneType
            Start position of region (zero-based, inclusive). Example values:
            '25221400'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/variantannotations/search'

        return self.rest_query(
            method_type,
            endpoint,
            variantAnnotationSetId=variant_annotation_set_id,
            effects=effects,
            end=end,
            pageSize=page_size,
            pageToken=page_token,
            referenceId=reference_id,
            referenceName=reference_name,
            start=start,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_variants_search(self, end, reference_name, start,
                                   variant_set_id, call_set_ids=None,
                                   page_size=None, page_token=None,
                                   norm_func=brc.default_pandas):
        """
        Return variant call information in GA4GH format for a region on a
        reference sequence. POST method. Queries:
        `ga4gh/variants/search <https://rest.ensembl.org/documentation/info/gavariants>`_.

        Parameters
        ----------
        end : int
            End position of region (zero-based, exclusive). Example values:
            '25455087'.
        reference_name : str
            Reference sequence name. Example values: '22'.
        start : int
            Start position of region (zero-based, inclusive). Example values:
            '25455086'.
        variant_set_id : str
            Return variant data for specific variantSets. Example values: '1'.
        call_set_ids : str or NoneType, optional, default: NoneType
            Return variant data for specific callSets. Example values: '[
            1:NA19777 ]'.
        page_size : int or NoneType, optional, default: NoneType
            Number of variants to show per page.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/variants/search'

        return self.rest_query(
            method_type,
            endpoint,
            end=end,
            referenceName=reference_name,
            start=start,
            variantSetId=variant_set_id,
            callSetIds=call_set_ids,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_variantsets_search(self, dataset_id, page_size=None,
                                      page_token=None,
                                      norm_func=brc.default_pandas):
        """
        Return a list of variant sets in GA4GH format. POST method. Queries:
        `ga4gh/variantsets/search <https://rest.ensembl.org/documentation/info/gavariantset>`_.

        Parameters
        ----------
        dataset_id : str
            Return variantSets by dataSet Identifier. Example values:
            '6e340c4d1e333c7a676b1710d2e3953c'.
        page_size : int or NoneType, optional, default: NoneType
            Number of variantSets to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/variantsets/search'

        return self.rest_query(
            method_type,
            endpoint,
            datasetId=dataset_id,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_variantsets(self, query_id, norm_func=brc.default_pandas):
        """
        Return the GA4GH record for a specific VariantSet given its identifier.
        GET method. Queries:
        `ga4gh/variantsets/:id <https://rest.ensembl.org/documentation/info/gavariantset_id>`_.

        Parameters
        ----------
        query_id : str
            VariantSet id. Example values: '1'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/variantsets/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_references_search(self, reference_set_id, accession=None,
                                     md5checksum=None, page_size=None,
                                     page_token=None,
                                     norm_func=brc.default_pandas):
        """
        Return a list of reference sequences in GA4GH format. POST method.
        Queries:
        `ga4gh/references/search <https://rest.ensembl.org/documentation/info/references>`_.

        Parameters
        ----------
        reference_set_id : str
            Return references for a referenceSet. Example values: 'GRCh38'.
        accession : str or NoneType, optional, default: NoneType
            Return reference information for a specific accession. Example
            values: 'NC_000021.9'.
        md5checksum : str or NoneType, optional, default: NoneType
            Return reference information for the md5checksum of the sequence.
            Example values: '9489ae7581e14efcad134f02afafe26c'.
        page_size : int or NoneType, optional, default: NoneType
            Number of references to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/references/search'

        return self.rest_query(
            method_type,
            endpoint,
            referenceSetId=reference_set_id,
            accession=accession,
            md5checksum=md5checksum,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_references(self, query_id, norm_func=brc.default_pandas):
        """
        Return data for a specific reference in GA4GH format by id. GET method.
        Queries:
        `ga4gh/references/:id <https://rest.ensembl.org/documentation/info/references_id>`_.

        Parameters
        ----------
        query_id : str
            Reference id. Example values: '9489ae7581e14efcad134f02afafe26c'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/references/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_referencesets_search(self, accession=None, page_size=None,
                                        page_token=None,
                                        norm_func=brc.default_pandas):
        """
        Return a list of reference sets in GA4GH format. POST method. Queries:
        `ga4gh/referencesets/search <https://rest.ensembl.org/documentation/info/referenceSets>`_.

        Parameters
        ----------
        accession : str or NoneType, optional, default: NoneType
            Return referenceSet information for a specific accession. Example
            values: '1'.
        page_size : int or NoneType, optional, default: NoneType
            Number of referenceSets to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/referencesets/search'

        return self.rest_query(
            method_type,
            endpoint,
            accession=accession,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_referencesets(self, query_id, norm_func=brc.default_pandas):
        """
        Return data for a specific reference set in GA4GH format. GET method.
        Queries:
        `ga4gh/referencesets/:id <https://rest.ensembl.org/documentation/info/referenceSets_id>`_.

        Parameters
        ----------
        query_id : str
            Reference set id.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/referencesets/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_ga4gh_variantannotationsets_search(self, variant_set_id,
                                                page_size=None,
                                                page_token=None,
                                                norm_func=brc.default_pandas):
        """
        Return a list of annotation sets in GA4GH format. POST method. Queries:
        `ga4gh/variantannotationsets/search <https://rest.ensembl.org/documentation/info/VariantAnnotationSet>`_.

        Parameters
        ----------
        variant_set_id : str
            Return variant annotation sets for a specific variant set. Example
            values: '1'.
        page_size : int or NoneType, optional, default: NoneType
            Number of variant annotation sets to return per request.
        page_token : int or NoneType, optional, default: NoneType
            Identifier showing which page of data to retrieve next.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        endpoint = 'ga4gh/variantannotationsets/search'

        return self.rest_query(
            method_type,
            endpoint,
            variantSetId=variant_set_id,
            pageSize=page_size,
            pageToken=page_token,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ga4gh_variantannotationsets(self, query_id,
                                        norm_func=brc.default_pandas):
        """
        Return meta data for a specific annotation set in GA4GH format. GET
        method. Queries:
        `ga4gh/variantannotationsets/:id <https://rest.ensembl.org/documentation/info/VariantAnnotationSet_id>`_.

        Parameters
        ----------
        query_id : str
            VariantAnnotation set id. Example values: 'Ensembl'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ga4gh/variantannotationsets/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class VariationGa4Gh(brc._BaseRestClient, _VariationGa4Gh):
    """Perform VariationGa4Gh queries
    """
    pass



"""Can be used as the main entry point for all the individual clients combined
into a single class (for ease of use)
"""
from ensembl_rest_client.base import base_rest_client as brc
from ensembl_rest_client import (
    archive as a,
    comp_genomics as c,
    eqtl as e,
    info as i,
    ld,
    lookup as lu,
    mapping as m,
    ontology as o,
    overlap as ov,
    phenotypes as ph,
    regulation as reg,
    sequence as seq,
    trans_haplotypes as trh,
    variation as v,
    variation_ga4gh as vg,
    vep,
    xrefs
)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Rest(a._Archive, c._ComparativeGenomics, e._Eqtl,
           i._Information, ld._LinkageDisequilibrium, lu._Lookup, m._Mapping,
           o._OntologiesAndTaxonomy, ov._Overlap, ph._PhenotypeAnnotations,
           reg._Regulation, seq._Sequence, trh._TranscriptHaplotypes,
           v._Variation, vg._VariationGa4Gh, vep._Vep, xrefs._CrossReferences,
           brc._BaseRestClient):
    """A single REST client combining all the rest client mix-ins.
    """
    pass



#!/usr/bin/env python
"""
Common functions that are used by many of the scripts
"""
from selenium.webdriver.chrome.options import Options
from selenium import webdriver
from collections import namedtuple
import inflection
import os
import re
import time
import pickle
import textwrap
import csv
import pprint as pp

MAX_WIDTH = 79
INDENT_CHARS = "    "

EndpointLink = namedtuple(
    'EndpointLink',
    ['name', 'key', 'endpoint', 'url', 'description']
)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def initialise_chrome(driver_path=None):
    """
    Initialise a chrome webdriver to download the files. Note that I can't make
    this headless at the moment as downloading does not seem to work in
    headless mode

    driver :str
        The path to the chrome webdriver (if it is not in your path)
    """

    chrome_options = Options()
    chrome_options.add_argument("--disable-extensions")
    chrome_options.add_argument("--disable-gpu")

    kwargs = {'options': chrome_options}
    if driver_path is not None:
        kwargs['executable_path'] = driver_path

    driver = webdriver.Chrome(**kwargs)

    return driver


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_endpoint_links(driver):
    """
    Get the links to all the endpoint webpages

    Parameters
    ----------
    driver : :obj:`selenium.Webdriver`
        The selenium webdriver

    Returns
    -------
    all_links : dict of list
        A each element of the list is an endpoint method
    """
    # Goto the rest landing page
    root_url = "http://rest.ensembl.org"
    driver.get(root_url)

    # All the links are in a big table
    main_table = driver.find_element_by_xpath(
        '//table[@class="table table-striped"]'
    )

    # Get the headings and the links associated with them
    headings = main_table.find_elements_by_xpath(
        './thead/tr/td/h3'
    )
    links = main_table.find_elements_by_xpath(
        './tbody'
    )

    # There should be the same amount of links and headings
    if len(links) != len(headings):
        raise ValueError("unmatched links and headings")

    all_links = {}
    for heading, link in zip(headings, links):
        heading_key = norm_string(heading.text)
        all_links[heading_key] = []
        for tr in link.find_elements_by_xpath('./tr'):
            link_data = [heading.text, heading_key]
            for idx, td in enumerate(tr.find_elements_by_xpath('./td')):
                if idx == 0:
                    link_data.append(td.text)
                    a = td.find_element_by_xpath('./a')
                    link_data.append(a.get_attribute('href'))
                else:
                    link_data.append(td.text)
            all_links[heading_key].append(EndpointLink(*link_data))
    return all_links


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def norm_string(value):
    """
    Normalise a string by making it lower case and replacing the spaces with
    underscores

    Parameters
    ----------
    value : str
        The string to normalise

    Returns
    -------
    norm_str
        The normalised string
    """
    return re.sub(r'\s+', '_', value.lower())


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_table_after(element):
    """
    Get the "sibling" table located after the element

    Parameters
    ----------
    element : :obj:`selenium.WebElement`
        The key element located before the table

    Returns
    -------
    table_element :  : :obj:`selenium.WebElement`
        The table element located after element, this will have the tag_name
        table
    """
    curr_tag = element.text
    next_table = element.find_element_by_xpath(
        './following-sibling::table'
    )
    back_check = next_table.find_element_by_xpath(
        './preceding-sibling::{0}'.format(element.tag_name)
    )

    # Make sure the table is the table that we expect
    if back_check.text != curr_tag:
        raise KeyError(
            "current: {0}, back check:{1}".format(
                curr_tag, back_check.text)
        )
    return next_table


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_resources(element):
    """
    Scrape the resources table. This has general information like return
    methods, query type GET/POST and max post size

    Parameters
    ----------
    element : :obj:`selenium.WebElement`
        The key element located before the table

    Returns
    -------
    resources : dict
        The keys are the resource name and the values are the resource value
    """
    resources = {}
    for tr in element.find_elements_by_xpath('./tbody/tr'):
        # print(tr)
        tds = tr.find_elements_by_xpath('./td')
        resources[norm_string(tds[0].text)] = tds[1].text

    try:
        resources['response_formats'] = \
            resources['response_formats'].split("\n")
    except KeyError:
        pass

    return resources


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_params(element):
    """
    Scrape a parameter section, these will be either an optional or required
    parameters

    Parameters
    ----------
    element : :obj:`selenium.WebElement`
        The key element located before the table

    Returns
    -------
    params : dict
        The keys name, type, desc, default, example_values
    """
    params = []
    for tr in element.find_elements_by_xpath('./tbody/tr'):
        param = {}
        tds = tr.find_elements_by_xpath('./td')

        param['name'] = tds[0].text
        param['type'] = tds[1].text
        param['desc'] = tds[2].text
        param['default'] = tds[3].text
        param['example_values'] = tds[4].text
        params.append(param)
    return params


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def parse_post_format(format_str):
    """
    """
    format_str = format_str.strip()

    if not re.match(r'^{.+?}$', format_str):
        if format_str == '':
            return {}

        raise TypeError(
            "unrecognised post format string: {0}".format(format_str)
        )
    elif re.match(r'^{\s*}$', format_str):
        return {}

    format_str = format_str.strip('{').strip('}')
    format_str = re.sub(r'[\s"]+', '', format_str)
    # format_str = re.split(r':,', format_str)

    format_str = re.sub(r'\[', '"ARRAY(', format_str)
    format_str = re.sub(r'\]', ')"', format_str)

    # I can't work out why this does not work??
    # quote = re.compile(
    #     r'''
    #     ,
    #     (?=
    #     (?:
    #     [^"]
    #     |
    #     "[^"]*"
    #     )*
    #     $)
    #     ''',
    #     re.MULTILINE
    # )
    # but this does??
    quote = re.compile(
        r''',(?=(?:[^'"`|]|'[^']*'|"[^"]*"|`[^`]*`|\|[^|]*\|)*$)'''
    )
    dict_delim = re.compile(
        r''':(?=(?:[^'"`|]|'[^']*'|"[^"]*"|`[^`]*`|\|[^|]*\|)*$)'''
    )

    format_str = quote.split(format_str)

    post_format = {}
    for i in format_str:
        key, value = dict_delim.split(i)
        # print('Key={0}'.format(key))
        # print('Value={0}'.format(value))
        if value == 'array':
            value = 'list'
        elif re.match(r'int(eger)?', value, re.IGNORECASE):
            value = 'int'
        elif value == 'long':
            value = 'int'
        elif value == 'float':
            value = 'float'
        elif value == 'string':
            value = 'str'
        elif value.startswith('"ARRAY('):
            value = 'list'
        else:
            for t in [int, float, str]:
                try:
                    value = type(t(value)).__name__
                    break
                except (ValueError, TypeError):
                    # raise
                    pass
        post_format[key] = value
    return post_format


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_post_message(element):
    """
    Scrape the resources table. This has general information like return
    methods, query type GET/POST and max post size

    Parameters
    ----------
    element : :obj:`selenium.WebElement`
        The key element located before the table

    Returns
    -------
    resources : dict
        The keys are the resource name and the values are the resource value
    """
    rows = element.find_elements_by_xpath('./tbody/tr')

    if len(rows) != 2:
        raise ValueError("unexpected row length: {0}".format(len(rows)))

    cols = rows[1].find_elements_by_xpath('./td')
    if len(cols) != 3:
        raise ValueError("unexpected col length: {0}".format(len(cols)))

    content_type = cols[0].text
    post_format = parse_post_format(cols[1].text)
    post_example = parse_post_format(cols[2].text)

    return {
        'content_type': content_type,
        'post_format': post_format,
        'post_example': post_example
    }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_link_params(driver, link):
    """
    Scrape all the resource and parameter data for the link.

    Parameters
    ----------
    driver : :obj:`selenium.Webdriver`
        The selenium webdriver
    link : :obj:`ElementLink`
        The link to extract

    Returns
    -------
    all_params : dict
        The keys are resources, required, optional and optionally post_params
    """
    driver.get(link.url)

    # Get the resources
    resource = driver.find_element_by_xpath(
        '//h3[text() = "Resource Information"]'
    )
    resource_table = get_table_after(resource)
    resources = get_resources(resource_table)

    # Get the required parameters
    required = driver.find_element_by_xpath(
        '//h3[text() = "Required"]'
    )
    required_table = get_table_after(required)
    required_params = get_params(required_table)

    # Get the optional parameters
    optional = driver.find_element_by_xpath(
        '//h3[text() = "Optional"]'
    )
    optional_table = get_table_after(optional)
    optional_params = get_params(optional_table)

    time.sleep(2)
    # Get the required parameters
    try:
        message = driver.find_element_by_xpath(
            '//h2[text() = "Message"]'
        )
        message_table = get_table_after(message)
        message_params = get_post_message(message_table)

        return {
            'resources': resources,
            'required': required_params,
            'optional': optional_params,
            'post_options': message_params
        }
    except Exception:
        if resources['methods'] == 'POST':
            raise
        return {
            'resources': resources,
            'required': required_params,
            'optional': optional_params
        }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def scrape_rest_attr(driver, links):
    """
    Parameters
    ----------
    driver : :obj:`selenium.Webdriver`
        The selenium webdriver
    links : dict of list
        Each element of the list is an `ElementLink`

    Returns
    -------
    rest_data : dict
        The keys are main rest endpoints and the values are dicts
    """
    rest_data = {}
    for key, values in links.items():
        rest_data[key] = []
        for i in values:
            time.sleep(1)
            params = get_link_params(driver, i)
            params['link'] = i
            rest_data[key].append(params)
        # break
    return rest_data


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def indent(level=1, value=""):
    """
    Indent a value by level

    Parameters
    ----------
    level : int, optional, default: 1
        The number of levels to indent
    value : str, optional, default: ""
        The value string to indent

    Returns
    -------
    indented_value : str
        The indented value
    """
    indent = INDENT_CHARS * level
    return "{0}{1}".format(indent, value)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def n_indent(level):
    """
    Return the number of indent characters for the required level

    Returns
    -------
    n_chars : int
        The number of indent spaces
    """
    return level * len(INDENT_CHARS)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_class_name(link_key):
    """
    Turn a link_key (rest "area") into a class name, this involved removing
    all underscores and making the first letter of all words uppercase

    Parameters
    ----------
    link_key : str
        The rest "area" that will correspond to a class name

    Returns
    -------
    class_name : str
        A camel-cased version of the link_key
    """
    return re.sub(r' ', '', re.sub(r'_', ' ', link_key).title())


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def write_docstring_quote(writer, level):
    """
    Write a quotation open/close for a docstring. The quote will have a
    carriage return after it

    Parameters
    ----------
    writer : :obj:`File`
        A file writer
    level : int
        The indent level
    """
    writer.write("{0}\n".format(indent(level, r'"""')))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def write_class(writer, link_key):
    """
    Write the class definition line

    Parameters
    ----------
    writer : :obj:`File`
        A file writer
    link_key : str
        The rest "area" that will correspond to a class name
    """
    class_name = get_class_name(link_key)
    writer.write(
        "{0}\nclass {1}(_BaseRestClient):\n".format(
            "\n\n# {0}".format("@" * 77),
            class_name
        )
    )
    write_docstring_quote(writer, 1)
    writer.write(indent(1, "Perform {0} queries\n".format(class_name)))
    write_docstring_quote(writer, 1)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_method_name(link):
    """
    Using the link endpoint, create a method name, by removing all parameter
    locations, converting the / and spaces into underscores and making
    lowercase

    Parameters
    ----------
    link : :obj:`EndpointLink`
        The scraped endpoint link

    Returns
    -------
    method_name : str
        The method name derived from the REST endpoint
    """
    name = re.sub(r':\w+', '', link.endpoint)
    name = re.sub(r'[/\s]+', '_', name)
    print(name)
    name = re.sub(r'(\W|_)+$', '', name)
    print(name)
    return name.lower()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def write_separator_line(writer, level):
    """
    Write a method separator line:
    # ~~~~~~~~~~~~~~~~~~~~~~~~

    Parameters
    ----------
    writer : :obj:`File`
        A file writer
    level : int
        The indent level
    """
    lead_chars = '# '
    line_len = MAX_WIDTH - (len(lead_chars) + n_indent(level))
    sep_line = "\n{0}".format(
        indent(level, "{0}{1}\n".format(lead_chars, '~' * line_len))
    )
    writer.write(sep_line)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def sub_python_reserved(name):
    """
    If the name collides with a python function call then rename
    """
    if name == 'id':
        return 'query_id'
    elif name == 'format':
        return 'data_format'
    elif name == 'filter':
        return 'query_filter'
    elif name == 'type':
        return 'data_type'
    elif name == 'type':
        return 'data_type'
    elif name == 'class':
        return 'class_method'
    else:
        return name


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def build_arg_list(arg_list, add_self=False):
    """
    Build an argument or kw argument list for a method call

    Parameters
    ----------
    arg_list : list of dict
        All the arguments, these may be option or required arguments
    add_self : bool, optional, default: False
        Add a "self" argument to the start of the list

    Returns
    -------
    arg_list : list or str
        A list of the required/optional arguments to the method
    """
    args = []
    if add_self is True:
        args.append('self')

    for i in arg_list:
        # Some arguments collide with python core functions, so rename any that
        # do in a standard way
        i['code_name'] = sub_python_reserved(i['name'])

        # Now we want to make any camel-cased arguments more pythonic
        i['code_name'] = inflection.underscore(i['code_name'])

        # We are skipping callbacks at the moment as I am not sure how to use
        # them
        if i['code_name'] == 'callback':
            continue
        args.append(i['code_name'])
    return args


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def add_post_options(method_def):
    """
    Sometimes not all the post options are defined in the required/optional
    tables. This double checks them and makes sure that they are defined.
    If they are not defined they are placed in args  (this may change when I
    can test everything)

    Parameters
    ----------
    method_def : dict
        The method definition dictionary having the keys: resources, link,
        required, optional
    """
    # Merge the post options that have been extracted from the format and
    # examples sections, with the format overiding the examples
    post_options = {**method_def['post_options']['post_example'],
                    **method_def['post_options']['post_format']}

    for arg, dtype in post_options.items():
        # The default is that it has not been found
        found = False

        # Now first see if we already have it
        for i in method_def['required']+method_def['optional']:
            if i['name'] == arg:
                found = True
                break

        # If we can't find the post arg in any of the required/optionals
        # then we add it
        if found is False:
            method_def['required'].append(
                dict(
                    name=arg,
                    type=dtype,
                    default='-',
                    example_values='-',
                    desc='Post option not defined in required/optional args'
                )
            )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def write_method_def(writer, method_def):
    """
    Write the method definition line

    Parameters
    ----------
    writer : :obj:`File`
        A file writer
    method_def : dict
        The method definition dictionary having the keys: resources, link,
        required, optional
    """
    # pp.pprint(method_def)
    call_def = 'def'
    method_name = get_method_name(method_def['link'])
    print("M =", method_name)
    subsequent_indent = len(call_def) + len(method_name) + 2 + n_indent(1)

    tr = textwrap.TextWrapper(
        width=MAX_WIDTH,
        initial_indent=indent(1),
        subsequent_indent=(' ' * subsequent_indent)
    )

    args = build_arg_list(method_def['required'], add_self=True)
    kwargs = build_arg_list(method_def['optional'], add_self=False)

    arg_str = ", ".join(args)
    kwarg_str = ", ".join(["{0}=None".format(k) for k in kwargs])

    all_args = []
    if len(arg_str) > 1:
        all_args.append(arg_str)

    if len(kwarg_str) > 0:
        all_args.append(kwarg_str)

    wrapped = tr.wrap(
        "{0} {1}({2}):\n".format(
            call_def,
            method_name,
            ", ".join(all_args)
        )
    )
    writer.write("{0}\n".format("\n".join(wrapped)))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_datatype(param_def):
    """
    Conver the REST "datatype" name into a python data type name

    Parameters
    ----------
    param_def : dict
        The information on the rest parameter

    Returns
    -------
    python_data_type_name : str
        The equivalent python data type name
    """
    rest_dt = param_def['type'].lower()
    allowed_values = ""

    if rest_dt == 'string':
        return 'str', allowed_values
    elif re.match(r'enum\((.+?)\)', rest_dt):
        values_regexp = re.match(r'enum\((.+?)\)', rest_dt)
        allowed_values = " Allowed values: {0}".format(
            values_regexp.group(1)
        )
        return 'str', allowed_values
    elif rest_dt.startswith('boolean'):
        return 'bool', allowed_values
    elif rest_dt == 'integer':
        return 'int', allowed_values
    elif rest_dt == '':
        return 'str', allowed_values
    else:
        return rest_dt, allowed_values


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_example_value_str(example):
    """
    Return a sensible example value string. If there is not one then return
    ""

    Parameters
    ----------
    example : str
        The scraped example value

    Returns
    -------
    example_str : str
        An example value string to place in the docstring
    """
    if example == '' or example == '-' or example == 'randomlygeneratedname':
        return ""
    else:
        return " Example values: '{0}'.".format(example)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_docstring_args(args, level, optional=False):
    """
    Generate a list of docstring arguments from the given arguments. This
    assumes that write_method_def has been called first.

    Parameters
    ----------
    args : list of dict
        The arguments to build the doctrings
    level : int
        The level the docstring will be written at
    optional : bool, optional, default: False
        Are the supplied arguments optional arguments, if so an optional
        string will be added to the doctring

    Returns
    -------
    docstring_args : list of str
        The docstring arguments
    """
    tr = textwrap.TextWrapper(
        width=MAX_WIDTH,
        initial_indent=indent(level+1),
        subsequent_indent=indent(level+1),
        break_long_words=False
    )

    optional_str = ""
    if optional is True:
        optional_str = " or NoneType, optional, default: NoneType"

    docstr_params = []
    for i in args:
        if i['code_name'] == 'callback':
            continue

        dtype, allowed_values = get_datatype(i)
        docstr_params.append(
            '{2}{0} : {1}{3}'.format(
                i['code_name'],
                dtype,
                indent(level),
                optional_str
            )
        )

        if not i['desc'].endswith('.'):
            i['desc'] = '{0}.'.format(i['desc'])

        # May need to adjust boolean example values here
        docstr_params.extend(
            tr.wrap(
                "{0}{1}{2}".format(
                    i['desc'],
                    get_example_value_str(i['example_values']),
                    allowed_values
                )
            )
        )
    return docstr_params


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def write_method_docstring(writer, method_def):
    """
    Write the method docstring line

    Parameters
    ----------
    writer : :obj:`File`
        A file writer
    method_def : dict
        The method definition dictionary having the keys: resources, link,
        required, optional
    """
    # pp.pprint(method_def)

    tr = textwrap.TextWrapper(
        width=MAX_WIDTH,
        initial_indent=indent(2),
        subsequent_indent=indent(2),
        break_long_words=False
    )

    docsting_quote = indent(2, '"""')
    desc = "{0}. {1} method. Queries [{2}]({3}).".format(
        re.sub(r'\.$', '', method_def['link'].description),
        method_def['resources']['methods'],
        re.sub(r'(GET|POST)\s+', '', method_def['link'].endpoint),
        method_def['link'].url
    )

    for msg, key in [('Max post size', 'maximum_post_size'),
                     ('Max slice length', 'slice_length')]:
        try:
            desc = '{0} {2}: {1}.'.format(
                desc,
                method_def['resources'][key],
                msg
            )
        except KeyError:
            pass
    desc = "\n".join(tr.wrap(desc))

    docstr_params = get_docstring_args(
        method_def['required'], 2, optional=False
    )
    docstr_params.extend(
        get_docstring_args(
            method_def['optional'], 2, optional=True
        )
    )

    if len(docstr_params) > 0:
        writer.write(
            "{0}\n{1}\n\n{2}Parameters\n{2}----------\n{3}\n{0}\n".format(
                docsting_quote,
                desc,
                indent(2),
                "\n".join(docstr_params)
            )
        )
    else:
        writer.write(
            "{0}\n{1}\n{0}\n".format(
                docsting_quote,
                desc
            )
        )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_endpoint(method_def):
    """
    """
    level = 2
    endpoint = re.sub(r'(GET|POST)\s*', '', method_def['link'].endpoint)

    params = []
    consumed = []
    for i in re.findall(r'(:\w+:?/|:\w+:?$)', endpoint):
        i = i.rstrip('/')
        var_name = i.strip(':')
        endpoint = re.sub(
            r'{0}'.format(i),
            '{{{0}}}'.format(var_name),
            endpoint
        )
        # Now find the code_name in the args
        found = False
        for arg in method_def['required']:
            if arg['name'] == var_name:
                consumed.append(arg['code_name'])
                params.append(
                    indent(
                        level+1,
                        value='{0}={1}'.format(var_name, arg['code_name'])
                    )
                )
                found = True
                break
        if found is False:
            return [indent(
                level,
                "raise NotImplementedError('this method call is broken')"
            )], consumed

    code = []
    if len(params) > 0:
        code.append(
            indent(level, value="endpoint = '{0}'.format(".format(endpoint))
        )
        code.append(',\n'.join(params))
        code.append(indent(level, value=")"))
    else:
        code.append(
            indent(level, value="endpoint = '{0}'".format(endpoint))
        )

    return code, consumed


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def write_code(writer, method_def):
    """
    Write the code for the REST method

    Parameters
    ----------
    writer : :obj:`File`
        A file writer
    method_def : dict
        The method definition dictionary having the keys: resources, link,
        required, optional
    """
    level = 2
    code = [indent(
        level,
        value="method_type = '{0}'".format(method_def['resources']['methods'])
        )]

    call_kwargs = [
        indent(level+1, 'method_type'),
        indent(level+1, 'endpoint')
    ]
    for var_name in ['maximum_post_size', 'slice_length']:
        try:
            value = method_def['resources'][var_name]
            call_kwargs.append(
                indent(level+1, '{0}={1}'.format(var_name, value))
                )
            code.append(
                indent(
                    level, value="{0} = {1}".format(var_name, value)
                )
            )
        except KeyError:
            pass

    endpoint, consumed = get_endpoint(method_def)
    code.extend(endpoint)
    code.append("")

    for i in method_def['required']+method_def['optional']:
        if i['code_name'] =='callback' or i['code_name'] in consumed:
            continue

        call_kwargs.append(
            indent(level+1, '{0}={1}'.format(i['name'], i['code_name']))
        )
    writer.write("\n".join(code))
    writer.write("\n")
    writer.write(
        "{0}return self.rest_query(\n{1}\n{0})\n".format(
            indent(level),
            ",\n".join(call_kwargs)
        )
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def write_method(writer, method_def):
    """
    Write a method definition to the module

    Parameters
    ----------
    writer : :obj:`File`
        A file writer
    method_def : dict
        The method definition dictionary having the keys: resources, link,
        required, optional
    """
    # We only output the method if it can respond in json format
    if 'json' not in method_def['resources']['response_formats']:
        return

    # The method separator line
    write_separator_line(writer, 1)

    # Now write the method definition
    write_method_def(writer, method_def)

    # Write the method docstring
    write_method_docstring(writer, method_def)
    write_code(writer, method_def)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
if __name__ == '__main__':
    driver = initialise_chrome(
        driver_path=os.path.join(
            os.environ['HOME'], 'Software/bin/chromedriver'
        ),
    )
    links = get_endpoint_links(driver)
    rest_data = scrape_rest_attr(driver, links)
    pickle.dump(rest_data, open('rest_params.p', 'wb'))
    rest_data = pickle.load(open('rest_params.p', 'rb'))
    # pp.pprint(rest_data)

    # with open('ensembl_rest_client.py', 'wt') as module_writer:
    #     for query_type, methods in rest_data.items():
    #         write_class(module_writer, query_type)
    #         print(query_type)
    #         for m in methods:
    #             try:
    #                 # Now we want to add in any post features that were not
    #                 # in the required/optional args
    #                 add_post_options(m)
    #             except KeyError:
    #                 # Not a post method
    #                 pass

    #             write_method(module_writer, m)

    pp.pprint(rest_data['information'])



"""Perform queries against the eQTL endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Eqtl(object):
    """Perform Eqtl queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_eqtl_stable_id(self, species, stable_id, statistic=None,
                           tissue=None, variant_name=None,
                           norm_func=brc.default_pandas):
        """
        Returns the p-value for each SNP in a given gene (e.g.
        ENSG00000227232). GET method. Queries:
        `eqtl/stable_id/:species/:stable_id <https://rest.ensembl.org/documentation/info/species_id>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        stable_id : str
            Ensembl stable ID. Example values: 'ENSG00000122435'.
        statistic : str or NoneType, optional, default: NoneType
            Filter by statistic. Example values: 'p-value beta'.
        tissue : str or NoneType, optional, default: NoneType
            Tissue of interest [Stomach, Thyroid, Whole_Blood]. Example values:
            'Whole_Blood'.
        variant_name : str or NoneType, optional, default: NoneType
            rsID (Reference SNP cluster ID). Example values: 'rs123'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'eqtl/stable_id/{species}/{stable_id}'.format(
            species=species,
            stable_id=stable_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            statistic=statistic,
            tissue=tissue,
            variant_name=variant_name,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_eqtl_variant_name(self, species, variant_name, query_id=None,
                              statistic=None, tissue=None,
                              norm_func=brc.default_pandas):
        """
        Returns the p-values for a SNP (e.g. rs123). GET method. Queries:
        `eqtl/variant_name/:species/:variant_name <https://rest.ensembl.org/documentation/info/species_variant>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        variant_name : str
            rsID (Reference SNP cluster ID). Example values: 'rs123'.
        query_id : str or NoneType, optional, default: NoneType
            Ensembl stable ID. Example values: 'ENSG00000122435'.
        statistic : str or NoneType, optional, default: NoneType
            Filter by statistic. Example values: 'p-value beta'.
        tissue : str or NoneType, optional, default: NoneType
            Tissue of interest [Stomach, Thyroid, Whole_Blood]. Example values:
            'Whole_Blood'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'eqtl/variant_name/{species}/{variant_name}'.format(
            species=species,
            variant_name=variant_name
        )

        return self.rest_query(
            method_type,
            endpoint,
            id=query_id,
            statistic=statistic,
            tissue=tissue,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_eqtl_tissue(self, species, norm_func=brc.default_pandas):
        """
        Returns all tissues currently available in the DB. GET method. Queries:
        `eqtl/tissue/:species/ <https://rest.ensembl.org/documentation/info/tissues>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'eqtl/tissue/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Eqtl(brc._BaseRestClient, _Eqtl):
    """Perform Eqtl queries
    """
    pass



from abc import ABC
from requests.exceptions import HTTPError
from ensembl_rest_client.base import cache as erc_cache
import requests
import json
import pandas as pd
import time
# import pprint as pp


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _BaseRestClient(ABC):
    """Base level Ensembl REST client, do not use directly.

    Parameters
    ----------
    url : `str`, optional, default: `https://rest.ensembl.org`
        The main URL were the endpoints are attached. Note the human
        genome build 37. Can be accessed at https://grch37.ensembl.org
        . However, please note that not all methods will work the
        GRCh37 version.
    pandas : `bool`, optional, default: `False`
        Return all data as `pandas.DataFrames`
    retries : `int`, optional, default: `2`
        The number of times to retry the query if it should fail with a
        potentially recoverable status code
    max_requests_per_sec : `int`, optional, default: `15`
        A positive integer. If the rate exceeds this then a pause is
        initiated to bring the rate into line
    ping: `bool`, optional, default: `True`
        Perform a ping when the object is instantiated. If the ping fails
        then an HTTPError will be raised
    cache : `str` or `bool`, optional, default: `False`
        Should caching be used, if `False`, then no cache is used. If `True`
        then a cache file will be created and used that corresponds to the
        endpoint URL. If it is a string, then that is used as a specific file
        to cache the data. Important, caches are not safe across multiple
        processes (and possibly threads) they are implemented with the Python
        sheve. If you are using multi-processing please supply a filename for
        each process. Note, that the cache will be used if the endpoint, kwargs
        and method type (post or get) are the same and was issued within the
        ``cache_seconds_cutoff``.
    cache_seconds_cutoff : `int`, optional, default `604800`
        The time in seconds that cache qeures must be within in order to be
        considered for use. The default is the number of seconds in 1 week.

    Raises
    ------
    HTTPError
        If the actual ping request fails or the ping returns a False value.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, url='https://rest.ensembl.org', pandas=False,
                 retries=2, max_requests_per_sec=15, ping=True,
                 cache=False, cache_seconds_cutoff=7 * erc_cache.DAY):
        # Make sure the URL has no trailing /
        self._url = url if not url.endswith('/') else url.rstrip('/')
        self.pandas = pandas
        self._max_requests_per_sec = max_requests_per_sec

        # Will hold the number of request that have been made after the
        # previous time check
        self._req_count = 0

        # Initialise the time as a starting point even though no requests have
        # been made yet
        self._last_req = time.time()
        self._retries = retries

        if self._retries < 0:
            raise ValueError("retries should be a positive integer or 0")

        if self._max_requests_per_sec < 1:
            raise ValueError(
                "max_requests_per_sec should be a positive integer"
            )

        # Get the cache and update/add method (or dummy if cache is not being
        # used)
        self.cache = erc_cache.get_cache(
            cache, seconds_cutoff=cache_seconds_cutoff
        )

        # Perform a ping check
        if ping is True:
            self.__class__.ping(url=self.url)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @classmethod
    def ping(cls, url='https://rest.ensembl.org'):
        """
        Ensure the Ensembl REST server is up before starting the test.

        Raises
        ------
        HTTPError
            If the actual ping request fails or the ping returns a False value
        """
        ping_endpoint = '{0}/info/ping?'.format(url)
        r = requests.get(
            ping_endpoint,
            headers={"Content-Type": "application/json"}
        )

        if not r.ok:
            r.raise_for_status()

        decoded = r.json()
        if bool(int(decoded['ping'])) is False:
            raise HTTPError(
                "The response was good but the ping was 0",
                response=r
            )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def rest_query(self, method_type, endpoint, **kwargs):
        """
        Perform a query against the Ensembl REST server

        Parameters
        ----------
        method_type : str
            The query type, can either be GET or POST
        endpoint : str
            The query endpoint, this will be joined with the root URL
        **kwargs
            Arguments specific to the query, if it is a post query, this should
            contain a `maximum_post_size` kwarg. If it is a region query, it
            should contain a `slice_length` kwarg. These are checked prior to
            querying the server. Other kwargs are specific for the query

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data

        Raises
        ------
        HTTPError
            If there is an unrecoverable error with the query or the number
            of tries is reached
        """
        # Make sure the endpoint has a starting / and join to the root URL
        # to create a query endpoint
        endpoint = endpoint if endpoint.startswith('/') \
            else '/{0}'.format(endpoint)
        query_endpoint = '{0}{1}'.format(self._url, endpoint)

        # Loop until we have reached the number of retries, the + 1 is the
        # initial query
        for i in range(self._retries + 1):
            # Check the cache
            cache_key = self.cache.get_key(
                (endpoint, method_type, kwargs)
            )
            cache_entry = self.cache.get_from_cache(cache_key)
            if cache_entry is not None:
                return cache_entry

            # Do we need to wait a bit before querying. It might make sense
            # to make this a class variable as multiple classes could be
            # querying
            self._check_rate_limit()

            try:
                if method_type == 'GET':
                    data = self._do_get_query(query_endpoint, **kwargs)
                    self.cache.add_to_cache(cache_key, data)
                    return data
                if method_type == 'POST':
                    data = self._do_post_query(query_endpoint, **kwargs)
                    self.cache.add_to_cache(cache_key, data)
                    return data
                else:
                    raise ValueError(
                        "unknown method type: '{0}'".format(method_type)
                    )
            except HTTPError as e:
                # If we get an error then we assess if we can recover from it
                # and retry
                if i == self._retries or \
                   self._assess_status_code(e.response) is False:
                    raise

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def url(self):
        """The rest base URL
        """
        return self._url

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _check_rate_limit(self):
        """
        """
        # check if we need to rate limit ourselves
        if self._req_count >= self._max_requests_per_sec:
            delta = time.time() - self._last_req
            if delta < 1:
                time.sleep(1 - delta)
            self._last_req = time.time()
            self._req_count = 0
        self._req_count += 1

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _assess_status_code(self, response):
        """
        Check the error status code and see if we can potentially recover from
        it

        Parameters
        ----------
        response : :obj:`requests.response`
            The request response object that has the status code

        Returns
        -------
        retry : bool
            If `True`, it means it is worth retrying the query if `False` then
            no we should just error out

        Raises
        ------
        ValueError
            If we get a status code that we do not know how to handle
        """
        if response.status_code == 400:
            return False
        elif response.status_code > 500:
            return False
        elif response.status_code == 404:
            return False
        elif response.status_code == 404:
            time.sleep(5)
            return True
        elif response.status_code == 408:
            time.sleep(5)
            return True
        elif response.status_code == 429:
            try:
                time.sleep(float(response.headers['retry-after']))
            except KeyError:
                time.sleep(5)
            return True
        else:
            raise ValueError(
                "unhandled response code: {0}".format(response.status_code)
            )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _do_get_query(self, endpoint, **kwargs):
        """
        Perform a GET query against the endpoint

        Parameters
        ----------
        endpoint : str
            A join of the base  URL and the specific endpoint
        **kwargs
            Arguments specific to the query

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        endpoint = '{0}?'.format(endpoint)
        kwargs, normalise = self._set_kwargs(kwargs)

        req_kwargs = dict(
            headers={"Content-Type": "application/json"}
        )
        if len(kwargs) > 0:
            req_kwargs['params'] = kwargs

        r = requests.get(endpoint, **req_kwargs)

        if not r.ok:
            r.raise_for_status()

        return self._return_data(r.json(), normalise, kwargs)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _do_post_query(self, endpoint, **kwargs):
        """
        Perform a GET query against the endpoint

        Parameters
        ----------
        endpoint : str
            A join of the base  URL and the specific endpoint
        **kwargs
            Arguments specific to the query

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        # pop out the non post kwargs, we will default it at 100, whatever
        # is left becomes post data
        max_post = kwargs.pop('maximum_post_size', 100)
        max_slice_len = kwargs.pop('slice_length', 1E6)
        # TODO: perform post/slice length/size checks

        kwargs, normalise = self._set_kwargs(kwargs)
        req_kwargs = dict(headers={"Content-Type": "application/json"})

        if len(kwargs) > 0:
            req_kwargs['data'] = json.dumps(kwargs)

        r = requests.post(endpoint, **req_kwargs)

        if not r.ok:
            r.raise_for_status()

        return self._return_data(r.json(), normalise, kwargs)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _set_kwargs(self, kwargs):
        """
        Remove any kwarg that is NoneType from the kwargs, as these will not
        have been defined by the user. TODO: Do I need to make ints/floats into
        strings or will requests do that? I probably need to convert bools to
        0/1 scale though

        Parameters
        ----------
        kwargs : dict
            The keyword arguments to process

        Returns
        -------
        kwargs : dict
            The processed kwargs, the return is not strictly necessary as these
            happen in place
        normalise : func
            A function to normalise the results data if pandas is True
        """
        # Extract the normalise function and fall back to the default pandas
        # one
        normalise = kwargs.pop('norm_func', default_pandas)

        keys = list(kwargs.keys())
        for k in keys:
            if kwargs[k] is None:
                del kwargs[k]

        return kwargs, normalise

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _return_data(self, data, normalise, kwargs):
        """
        Format the return data based on what the user wants

        Parameters
        ----------
        data : list or dict
            The data returned from the REST query to format
        normalise : func
            A normalisation function that is applied if pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        if self.pandas is True:
            return normalise(data, **kwargs)
        return data


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def default_pandas(data, **kwargs):
    """
    A default pandas normaliser, this does not try to handle any nesting
    and just drops the data in a `pandas.DataFrame`

    Parameters
    ----------
    data : list or dict
        The data to convert into a `pandas.DataFrame`
    **kwargs
        Any keyword arguments passed to the query that the normalise function
        will need to know about to do it's job

    Returns
    -------
    pandas_data : :obj:`pandas.DataFrame`
        The data shoehorned into a `pandas.DataFrame`
    """
    if isinstance(data, dict):
        data = [data]
    return pd.DataFrame(data)



"""Perform queries against the information endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Information(object):
    """Perform Information queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_analysis(self, species, norm_func=brc.default_pandas):
        """List the names of analyses involved in generating Ensembl data. GET
        method. Queries:
        `info/analysis/:species <https://rest.ensembl.org/documentation/info/analysis>`_.

        Parameters
        ----------
        species : `str`
            Species name/alias. Example values: 'homo_sapiens'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/analysis/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_assembly(self, species, bands=None, synonyms=None,
                          norm_func=brc.default_pandas):
        """List the currently available assemblies for a species, along with
        toplevel sequences, chromosomes and cytogenetic bands. GET method.
        Queries:
        `info/assembly/:species <https://rest.ensembl.org/documentation/info/assembly_info>`_.

        Parameters
        ----------
        species : `str`
            Species name/alias. Example values: 'homo_sapiens human'.
        bands : `bool` or `NoneType`, optional, default: `NoneType`
            If set to `True`, include karyotype band information. Only display
            if band information is available.
        synonyms : `bool` or `NoneType`, optional, default: `NoneType`
            If set to `True`, include information about known synonyms.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/assembly/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            bands=bands,
            synonyms=synonyms,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_region_assembly(self, region_name, species, bands=None,
                                 synonyms=None,  norm_func=brc.default_pandas):
        """Returns information about the specified toplevel sequence region for
        the given species. GET method. Queries:
        `info/assembly/:species/:region_name <https://rest.ensembl.org/documentation/info/assembly_stats>`_.

        Parameters
        ----------
        region_name : `str`
            The (top level) sequence region name. Example values: 'X'.
        species : `str`
            Species name/alias. Example values: 'homo_sapiens human'.
        bands : `bool` or `NoneType`, optional, default: `NoneType`
            If set to `True`, include karyotype band information. Only display
            if band information is available.
        synonyms : `bool` or `NoneType`, optional, default: `NoneType`
            If set to `True`, include information about known synonyms.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/assembly/{species}/{region_name}'.format(
            species=species,
            region_name=region_name
        )

        return self.rest_query(
            method_type,
            endpoint,
            bands=bands,
            synonyms=synonyms,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_biotypes(self, species, norm_func=brc.default_pandas):
        """List the functional classifications of gene models that Ensembl
        associates with a particular species. Useful for restricting the type
        of genes/transcripts retrieved by other endpoints. GET method. Queries:
        `info/biotypes/:species <https://rest.ensembl.org/documentation/info/biotypes>`_.

        Parameters
        ----------
        species : `str`
            Species name/alias. Example values: 'homo_sapiens'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/biotypes/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_biotypes_groups(self, group=None, object_type=None,
                                 norm_func=brc.default_pandas):
        """Without argument the list of available biotype groups is returned. With
        :group argument provided, list the properties of biotypes within that
        group. Object type (gene or transcript) can be provided for filtering.
        GET method. Queries:
        `info/biotypes/groups/:group/:object_type <https://rest.ensembl.org/documentation/info/biotypes_groups>`_.

        Parameters
        ----------
        group : `str`, optional, default: `NoneType`
            Biotype group. Example values: 'coding'.
        object_type : `str`, optional, default: `NoneType`
            Object type (gene or transcript). Example values: 'gene'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/biotypes/groups'

        # These are optional but are encoded in the URL endpoint directly
        # I think this differs from other optional arguments
        # TODO: I need a generic solution to this
        if group is not None:
            endpoint = '/{endpoint}/{group}'.format(
                endpoint=endpoint,
                group=group
            )

            if object_type is not None:
                endpoint = '/{endpoint}/{object_type}'.format(
                    endpoint=endpoint,
                    object_type=object_type
                )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_biotypes_name(self, name, object_type=None,
                               norm_func=brc.default_pandas):
        """List the properties of biotypes with a given name. Object type (gene
        or transcript) can be provided for filtering. GET method. Queries:
        `info/biotypes/name/:name/:object_type <https://rest.ensembl.org/documentation/info/biotypes_name>`_

        Parameters
        ----------
        name : `str`
            Biotype name. Example values: 'protein_coding'.
        object_type : `str`, optional, default: `NoneType`
            Object type (gene or transcript). Example values: 'gene'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/biotypes/name/{name}'.format(
            name=name
        )

        # These are optional but are encoded in the URL endpoint directly
        # I think this differs from other optional arguments
        # TODO: I need a generic solution to this
        if object_type is not None:
            endpoint = '/{endpoint}/{object_type}'.format(
                endpoint=endpoint,
                object_type=object_type
            )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_compara_methods(self, class_method=None, compara=None,
                                 norm_func=brc.default_pandas):
        """List all compara analyses available (an analysis defines the type of
        comparative data). GET method. Queries:
        `info/compara/methods <https://rest.ensembl.org/documentation/info/compara_methods>`_

        Parameters
        ----------
        class_method : `str` or `NoneType`, optional, default: `NoneType`
            The class of the method to query for. Regular expression patterns
            are supported. Example values: 'GenomicAlign'.
        compara : `str` or `NoneType`, optional, default: `NoneType`
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/compara/methods'

        kwargs = {
            'class': class_method,
            'compara': compara
        }
        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func,
            **kwargs
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_compara_species_sets(self, method, compara=None,
                                      norm_func=brc.default_pandas):
        """List all collections of species analysed with the specified compara
        method. GET method. Queries:
        `info/compara/species_sets/:method <https://rest.ensembl.org/documentation/info/compara_species_sets>`_

        Parameters
        ----------
        method : `str`
            Filter by compara method. Use one the methods returned by
            /info/compara/methods endpoint. Example values: 'EPO'.
        compara : `str` or `NoneType`, optional, default: `NoneType`
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/compara/species_sets/{method}'.format(
            method=method
        )

        return self.rest_query(
            method_type,
            endpoint,
            compara=compara,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_comparas(self, norm_func=brc.default_pandas):
        """Lists all available comparative genomics databases and their data
        release. DEPRECATED: use info/genomes/division instead. GET method.
        Queries:
        `info/comparas <https://rest.ensembl.org/documentation/info/comparas>`_

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/comparas'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_data(self, norm_func=brc.default_pandas):
        """Shows the data releases available on this REST server. May return more
        than one release (unfrequent non-standard Ensembl configuration). GET
        method. Queries:
        `info/data <https://rest.ensembl.org/documentation/info/data>`_

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/data'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_eg_version(self, norm_func=brc.default_pandas):
        """Returns the Ensembl Genomes version of the databases backing this
        service. GET method. Queries:
        `info/eg_version <https://rest.ensembl.org/documentation/info/eg_version>`_

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/eg_version'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_external_dbs(self, species, feature=None, query_filter=None,
                              norm_func=brc.default_pandas):
        """Lists all available external sources for a species. GET method. Queries:
        `info/external_dbs/:species <https://rest.ensembl.org/documentation/info/external_dbs>`_

        Parameters
        ----------
        species : `str`
            Species name/alias. Example values: 'homo_sapiens'.
        feature : `str` or `NoneType`, optional, default: `NoneType`
            Only return external DB entries for a given feature. Example
            values: 'xref dna_align_feature'. Allowed values: dna_align_feature
            protein_align_feature unmapped_object xref seq_region_synonym
        query_filter : `str` or `NoneType`, optional, default: `NoneType`
            Restrict external DB searches to a single source or pattern. SQL-
            LIKE patterns are supported. Example values: 'HGNC GO%'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/external_dbs/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            feature=feature,
            filter=query_filter,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_divisions(self, norm_func=brc.default_pandas):
        """Get list of all Ensembl divisions for which information is available.
        GET method. Queries:
        `info/divisions <https://rest.ensembl.org/documentation/info/info_divisions>`_

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/divisions'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_genomes(self, genome_name, expand=None,
                         norm_func=brc.default_pandas):
        """Find information about a given genome. GET method. Queries:
        `info/genomes/:genome_name <https://rest.ensembl.org/documentation/info/info_genome>`_

        Parameters
        ----------
        genome_name : `str`
            The production name of the genome. Example values:
            'nanoarchaeum_equitans_kin4_m'.
        expand : `bool` or `NoneType`, optional, default: `NoneType`
            Expands the information to include details of sequences. Can be
            very large.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/genomes/{genome_name}'.format(
            genome_name=genome_name
        )

        return self.rest_query(
            method_type,
            endpoint,
            expand=expand,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_genomes_accession(self, accession, expand=None,
                                   norm_func=brc.default_pandas):
        """Find information about genomes containing a specified INSDC accession.
        GET method. Queries:
        `info/genomes/accession/:accession <https://rest.ensembl.org/documentation/info/info_genomes_accession>`_

        Parameters
        ----------
        accession : `str`
            INSDC sequence accession (optionally versioned). Example values:
            'U00096'.
        expand : `bool` or `NoneType`, optional, default: `NoneType`
            Expands the information to include details of sequences. Can be
            very large.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/genomes/accession/{accession}'.format(
            accession=accession
        )

        return self.rest_query(
            method_type,
            endpoint,
            expand=expand,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_genomes_assembly(self, assembly_id, expand=None,
                                  norm_func=brc.default_pandas):
        """Find information about a genome with a specified assembly. GET method.
        Queries:
        `info/genomes/assembly/:assembly_id <https://rest.ensembl.org/documentation/info/info_genomes_assembly>`_

        Parameters
        ----------
        assembly_id : `str`
            INSDC assembly ID (optionally versioned). Example values:
            'GCA_000005005.6'.
        expand : `bool` or `NoneType`, optional, default: `NoneType`
            Expands the information to include details of sequences. Can be
            very large.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/genomes/assembly/{assembly_id}'.format(
            assembly_id=assembly_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            expand=expand,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_genomes_division(self, division, expand=None,
                                  norm_func=brc.default_pandas):
        """Find information about all genomes in a given division. May be large
        for Ensembl Bacteria. GET method. Queries:
        `info/genomes/division/:division_name <https://rest.ensembl.org/documentation/info/info_genomes_division>`_

        Parameters
        ----------
        division : `str`
            The name of the division. Example values: 'EnsemblPlants'.
        expand : `bool` or `NoneType`, optional, default: `NoneType`
            Expands the information to include details of sequences. Can be
            very large.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/genomes/division/{division}'.format(
            division=division
        )

        return self.rest_query(
            method_type,
            endpoint,
            expand=expand,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_genomes_taxonomy(self, taxon_name, expand=None,
                                  norm_func=brc.default_pandas):
        """Find information about all genomes beneath a given node of the
        taxonomy. GET method. Queries:
        `info/genomes/taxonomy/:taxon_name <https://rest.ensembl.org/documentation/info/info_genomes_taxonomy>`_

        Parameters
        ----------
        taxon_name : `str`
            Taxon name or NCBI taxonomy ID. Example values: 'Homo sapiens'.
        expand : `bool` or `NoneType`, optional, default: `NoneType`
            Expands the information to include details of sequences. Can be
            very large.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/genomes/taxonomy/{taxon_name}'.format(
            taxon_name=taxon_name
        )

        return self.rest_query(
            method_type,
            endpoint,
            expand=expand,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_ping(self, norm_func=brc.default_pandas):
        """Checks if the service is alive. GET method. Queries:
        `info/ping <https://rest.ensembl.org/documentation/info/ping>`_

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/ping'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_rest(self, norm_func=brc.default_pandas):
        """Shows the current version of the Ensembl REST API. GET method. Queries:
        `info/rest <https://rest.ensembl.org/documentation/info/rest>`_

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/rest'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_software(self, norm_func=brc.default_pandas):
        """Shows the current version of the Ensembl API used by the REST server.
        GET method. Queries:
        `info/software <https://rest.ensembl.org/documentation/info/software>`_

        Parameters
        ----------
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/software'

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_species(self, division=None, hide_strain_info=None,
                         strain_collection=None, norm_func=brc.default_pandas):
        """Lists all available species, their aliases, available adaptor groups
        and data release. GET method. Queries:
        `info/species <https://rest.ensembl.org/documentation/info/species>`_

        Parameters
        ----------
        division : `str` or `NoneType`, optional, default: `NoneType`
            Filter by Ensembl or Ensembl Genomes division. Example values:
            'EnsemblVertebrates'.
        hide_strain_info : `bool` or `NoneType`, optional, default: `NoneType`
            Show/hide strain and strain_collection info in the output.
        strain_collection : `str` or `NoneType`, optional, default: `NoneType`
            Filter by strain_collection. Example values: 'mouse'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/species'

        return self.rest_query(
            method_type,
            endpoint,
            division=division,
            hide_strain_info=hide_strain_info,
            strain_collection=strain_collection,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_variation(self, species, query_filter=None,
                           norm_func=brc.default_pandas):
        """List the variation sources used in Ensembl for a species. GET method.
        Queries:
        `info/variation/:species <https://rest.ensembl.org/documentation/info/variation>`_

        Parameters
        ----------
        species : `str`
            Species name/alias. Example values: 'homo_sapiens'.
        query_filter : `str` or `NoneType`, optional, default: `NoneType`
            Restrict the variation source searches to a single source. Example
            values: 'dbSNP ClinVar OMIM UniProt HGMD'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/variation/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            filter=query_filter,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_variation_consequence_types(self, rank=None,
                                             norm_func=brc.default_pandas):
        """Lists all variant consequence types. GET method. Queries:
        `info/variation/consequence_types <https://rest.ensembl.org/documentation/info/variation_consequence_types>`_

        Parameters
        ----------
        rank : `bool` or `NoneType`, optional, default: `NoneType`
            Include consequence ranking.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/variation/consequence_types'

        return self.rest_query(
            method_type,
            endpoint,
            rank=rank,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_variation_individuals(self, population_name, species,
                                       norm_func=brc.default_pandas):
        """List all individuals for a population from a species. GET method.
        Queries:
        `info/variation/populations/:species:/:population_name <https://rest.ensembl.org/documentation/info/variation_population_name>`_

        Parameters
        ----------
        population_name : `str`
            Population name. Example values: '1000GENOMES:phase_3:ASW'.
        species : `str`
            Species name/alias. Example values: 'human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = (
            'info/variation/populations/{species}/{population_name}'.format(
                species=species,
                population_name=population_name
            )
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_info_variation_populations(self, species, query_filter=None,
                                       norm_func=brc.default_pandas):
        """List all populations for a species. GET method. Queries:
        `info/variation/populations/:species <https://rest.ensembl.org/documentation/info/variation_populations>`_

        Parameters
        ----------
        species : `str`
            Species name/alias. Example values: 'homo_sapiens'.
        query_filter : `str` or `NoneType`, optional, default: `NoneType`
            Restrict populations returned to e.g. only populations with LD
            data. It is highly recommended to set a filter and to avoid loading
            the complete list of populations. Example values: 'LD'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'info/variation/populations/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            filter=query_filter,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Information(brc._BaseRestClient, _Information):
    """Perform Information queries
    """
    pass
